import importlib
import os
import pathlib
from typing import NamedTuple, Mapping

import toml

from gaql_console import exceptions


class GAQLContext(NamedTuple):
    client_customer_id: str
    client_id: str
    client_secret: str
    developer_token: str
    login_customer_id: str
    refresh_token: str


def ensure_full_context(candidate: Mapping, source: str) -> GAQLContext:
    required = set(GAQLContext._fields)
    available = set(candidate.keys())
    missing = required.difference(available)
    if missing:
        missing_str = ", ".join(missing)
        raise exceptions.ContextException(f"Following context values were missing in {source}: {missing_str}", missing)
    return GAQLContext(**{k: str(candidate[k]) for k in required})


def parse_pyproject_toml() -> dict:
    # @TODO: try more locations
    location = pathlib.Path("pyproject.toml")
    if not location.is_file():
        raise exceptions.ContextException(f"Unable to locate pyproject.toml under {location}")
    with location.open() as f:
        try:
            toml_data = toml.load(f)
            return toml_data["tool"]["gaql-console"]
        except KeyError:
            raise exceptions.ContextException("No [tool.gaql-console] section in pyproject.toml.")


def build_context(source: str, *args, **kwargs) -> GAQLContext:
    return CONTEXT_CHOICES[source](*args, **kwargs)


def from_anywhere(*args, **kwargs) -> GAQLContext:
    # @TODO: from argparse
    # @TODO: from pyproject.toml
    sources = (from_envvars, from_callable)
    inner_exceptions = []
    for source in sources:
        try:
            return source(*args, **kwargs)
        except exceptions.ContextException as exc:
            inner_exceptions.append(exc)
    sources_str = ", ".join(s.__name__ for s in sources)
    raise exceptions.ContextException(f"Unable to populate valid GAQLContext. Used {sources_str}.", inner_exceptions)


def from_envvars(*args, **kwargs) -> GAQLContext:
    prefix = "GAQL_"
    raw_context = {k.lstrip(prefix).lower(): os.environ[k] for k in os.environ if k.startswith(prefix)}
    return ensure_full_context(raw_context, "envvars")


def from_callable(*args, **kwargs) -> GAQLContext:
    toml_data = parse_pyproject_toml()
    context_callable = toml_data.get("context-callable")
    if not context_callable:
        raise exceptions.ContextException("No context-callable under [tool.gaql-console]")

    module_name, callable_name = context_callable.split(":", maxsplit=1)

    # @TODO: better exception handling
    module = importlib.import_module(module_name)
    callable = getattr(module, callable_name)

    raw_context = callable(*args, **kwargs)
    return ensure_full_context(raw_context, callable_name)


CONTEXT_CHOICES = {
    "anywhere": from_anywhere,
    "envvars": from_envvars,
    "callable": from_callable,
}



from typing import Iterable

from prompt_toolkit import completion, document
from prompt_toolkit.completion import word_completer
from pygments import lexer, token

FUNCTIONS = {
    "LAST_14_DAYS",
    "LAST_30_DAYS",
    "LAST_7_DAYS",
    "LAST_BUSINESS_WEEK",
    "LAST_MONTH",
    "LAST_WEEK_MON_SUN",
    "LAST_WEEK_SUN_SAT",
    "THIS_MONTH",
    "THIS_WEEK_MON_TODAY",
    "THIS_WEEK_SUN_TODAY",
    "TODAY",
    "YESTERDAY",
}
# @TOOD: PARAMETERS clause - what is this for
KEYWORDS = {"SELECT", "FROM", "WHERE", "ORDER BY", "LIMIT", "ASC", "DESC", "AND"}
OPERATORS = {
    "IN",
    "NOT IN",
    "LIKE",
    "NOT LIKE",
    "CONTAINS ANY",
    "CONTAINS ALL",
    "CONTAINS NONE",
    "IS NULL",
    "IS NOT NULL",
    "DURING",
    "BETWEEN",
    "REGEXP_MATCH",
    "NOT REGEXP_MATCH",
}


class GAQLLexer(lexer.RegexLexer):
    """
    See: https://developers.google.com/google-ads/api/docs/query/grammar
    """

    name = "GAQL"
    aliases = ["gaql"]
    filenames = ["*.gaql"]

    function_tokens = [(f"(?i)({t})", token.Keyword) for t in FUNCTIONS]
    keyword_tokens = [(f"(?i)({t})", token.Keyword) for t in KEYWORDS]
    operator_tokens = [(f"(?i)({t})", token.Operator) for t in OPERATORS]

    tokens = {
        "root": [
            (r"\s+", token.Whitespace),
            (r",", token.Punctuation),
            *keyword_tokens,
            (r"(=|!=|>|>=|<|<=)", token.Operator),
            *operator_tokens,
            *function_tokens,
            (r"([\w.]+)", token.Name.Attribute),
        ],
    }


gaql_autocompleter = word_completer.WordCompleter([*FUNCTIONS, *KEYWORDS, *OPERATORS], ignore_case=True)


class GAQLCompleter(completion.Completer):
    def _complete_with_rewind(self, document: document.Document, word: str) -> Iterable[completion.Completion]:
        words = document.text_before_cursor.rsplit(maxsplit=1)
        if not words:
            return

        final_word = words[-1].upper()

        len_final_word = len(final_word)
        if len_final_word == len(word):
            return
        if final_word == word[:len_final_word]:
            yield completion.Completion(word, -len_final_word)

    def get_completions(
        self, _document: document.Document, complete_event: completion.CompleteEvent
    ) -> Iterable[completion.Completion]:
        text_before = _document.text_before_cursor

        if "SELECT" not in text_before:
            yield from self._complete_with_rewind(_document, "SELECT")

        # FROM must come after any fields
        # fields have to have a dot in the lookup path
        if "." in text_before and "FROM" not in text_before:
            yield from self._complete_with_rewind(_document, "FROM")

        if "FROM" in text_before:
            yield from self._complete_with_rewind(_document, "WHERE")
            yield from self._complete_with_rewind(_document, "ORDER BY")
            yield from self._complete_with_rewind(_document, "LIMIT")



import importlib
from typing import Iterator

from google.ads.googleads import client as ads_client, errors as ads_errors
from google.oauth2 import credentials

from gaql_console import context, exceptions


# @TODO: Make me customizable
API_VERSION = "v9"


def import_for_version(suffix: str):
    module_path, member = suffix.rsplit(".", maxsplit=1)
    path = f"google.ads.googleads.{API_VERSION}.{module_path}"
    module = importlib.import_module(path)
    member = getattr(module, member)
    return member


GoogleAdsRow = import_for_version("services.types.google_ads_service.GoogleAdsRow")


class GAQLClient:
    _context: context.GAQLContext
    _client: ads_client.GoogleAdsClient

    def __init__(self, _context: context.GAQLContext):
        self._context = _context
        _credentials = credentials.Credentials.from_authorized_user_info(
            {
                "refresh_token": _context.refresh_token,
                "client_id": _context.client_id,
                "client_secret": _context.client_secret,
            }
        )
        self._client = ads_client.GoogleAdsClient(
            credentials=_credentials,
            developer_token=_context.developer_token,
            login_customer_id=_context.login_customer_id,
            version=API_VERSION,
        )

    def query(self, gaql: str) -> Iterator[GoogleAdsRow]:
        # keep this in scope - see https://github.com/googleads/google-ads-python/issues/384
        service = self._client.get_service("GoogleAdsService")

        try:
            streaming_response_iterator = service.search_stream(
                customer_id=self._context.client_customer_id, query=gaql
            )
        except ads_errors.GoogleAdsException as exc:
            raise exceptions.QueryException(exc.failure.errors[0].message)

        for response in streaming_response_iterator:
            for result in response.results:
                yield result


def client(_context: context.GAQLContext) -> ads_client.GoogleAdsClient:
    return ads_client.GoogleAdsClient(
        credentials=None,
        developer_token=_context.developer_token,
        version=API_VERSION,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gaql_console']

package_data = \
{'': ['*']}

install_requires = \
['Pygments>=2.11.2,<3.0.0',
 'click>=8.0.3,<9.0.0',
 'google-ads>=17.0.0,<18.0.0',
 'prompt-toolkit>=3.0.24,<4.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['gaql = gaql_console.console:main']}

setup_kwargs = {
    'name': 'gaql-console',
    'version': '0.1.1',
    'description': 'Google Ads Query Language Interactive Console',
    'long_description': None,
    'author': 'Piotr Kilczuk',
    'author_email': 'piotr@kilczuk.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import contextlib
from typing import List, Optional


@contextlib.contextmanager
def handle_console_errors():
    try:
        yield
    except EOFError:
        exit(0)
    except KeyboardInterrupt:
        exit(1)


class QueryException(Exception):
    message: str

    def __init__(self, message: str):
        self.message = message


class ContextException(ValueError):
    message: str
    inner_exceptions: List[ContextException]

    def __init__(self, message: str, inner_exceptions: Optional[List[Exception]] = None):
        self.message = message
        self.inner_exceptions = inner_exceptions or []
